from vg.compat import v2 as vg


class AnalysisMixin:
    @property
    def vertex_centroid(self):
        """
        The centroid or geometric average of the vertices.
        """
        return vg.average(self.v)

    @property
    def bounding_box(self):
        """
        A bounding box around the vertices.

        Returns:
            polliwog.Box: The bounding box.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Box
        """
        from polliwog import Box

        return Box.from_points(self.v)

    def apex(self, along):
        """
        Find the most extreme vertex in the direction provided.

        Args:
            along (np.arraylike): A `(3,)` direction of interest.

        Returns:
            np.ndarray: A copy of the point in `self.v` which lies furthest
                in the direction of interest.
        """
        return vg.apex(self.v, along=along)

    def face_normals(self, normalize=True):
        """
        Compute surface normals of each face. The direction of the normal
        follows conventional counter-clockwise winding and the right-hand rule.

        Args:
            normalize (bool): When True, return unit-length normals.

        Returns:
            np.ndarray: Face normals as `(k, 3)`.
        """
        from polliwog.tri import surface_normals

        return surface_normals(self.v[self.f], normalize=normalize)



import typing as t
import click

LengthUnits = t.Literal["m", "cm", "mm", "in"]
ALL_LENGTH_UNITS: t.List[LengthUnits] = ["m", "cm", "mm", "in"]


@click.command
@click.option("--outdir", type=click.Path(exists=True, dir_okay=True, file_okay=False))
@click.option("--verbose/--no-verbose", default=True)
@click.argument("from_units", type=click.Choice(ALL_LENGTH_UNITS), metavar="FROM_UNITS")
@click.argument("to_units", type=click.Choice(ALL_LENGTH_UNITS), metavar="TO_UNITS")
@click.argument("mesh_path", nargs=-1, type=click.Path(exists=True))
def convert_units(
    from_units: LengthUnits,
    to_units: LengthUnits,
    mesh_path: t.Tuple[str],
    outdir: t.Optional[str],
    verbose: bool,
):
    import os
    from lacecore import load_obj

    def pif(message: str) -> None:
        if verbose:
            click.echo(message, err=True)

    for this_mesh_path in mesh_path:
        if outdir is None:
            filename, extension = os.path.splitext(os.path.basename(this_mesh_path))
            output_path = f"{filename}_{to_units}{extension}"
        else:
            output_path = os.path.join(outdir, os.path.basename(this_mesh_path))

        pif(
            f"Converting {this_mesh_path} from {from_units} to {to_units}",
        )

        load_obj(this_mesh_path).units_converted(
            from_units=from_units, to_units=to_units
        ).write_obj(output_path)

        pif(f"  Wrote {output_path}")


if __name__ == "__main__":  # pragma: no cover
    convert_units()



from .selection_object import Selection


class SelectionMixin:
    def select(self):
        """
        Begin a chained selection operation. After invoking `.select()`,
        apply selection criteria, then invoke `.end()` to create a submesh.

        Include `.union()` in the chain to combine multiple sets of
        selection criteria into a single submesh.

        Does not mutate the callee.

        Returns:
            lacecore.Selection: The selection operation.

        Example:
            >>> centroid = np.average(mesh.v, axis=0)
            >>> upper_right_quadrant = (
                mesh.select()
                .vertices_above(centroid, dim=0)
                .vertices_above(centroid, dim=1)
                .end()
            )
            >>> upper_half_plus_right_half = (
                mesh.select()
                .vertices_above(centroid, dim=0)
                .union()
                .vertices_above(centroid, dim=1)
                .end()
            )
        """
        return Selection(target=self)

    def keeping_vertices_at_or_above(self, dim, point):
        """
        Select vertices which, when projected to the given axis, are either
        coincident with the projection of the given point, or lie further
        along the axis.

        Return a new mesh, without mutating the callee.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().vertices_at_or_above(dim=dim, point=point).end()

    def keeping_vertices_above(self, dim, point):
        """
        Select vertices which, when projected to the given axis, lie further
        along that axis than the projection of the given point.

        Return a new mesh, without mutating the callee.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().vertices_above(dim=dim, point=point).end()

    def keeping_vertices_at_or_below(self, dim, point):
        """
        Select vertices which, when projected to the given axis, are either
        coincident with the projection of the given point, or lie before it.

        Return a new mesh, without mutating the callee.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().vertices_at_or_below(dim=dim, point=point).end()

    def keeping_vertices_below(self, dim, point):
        """
        Select vertices which, when projected to the given axis, lie before
        the projection of the given point.

        Return a new mesh, without mutating the callee.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().vertices_below(dim=dim, point=point).end()

    def keeping_vertices_on_or_in_front_of_plane(self, plane):
        """
        Select the vertices which are either on or in front of the given
        plane.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        return self.select().vertices_on_or_in_front_of_plane(plane=plane).end()

    def keeping_vertices_in_front_of_plane(self, plane):
        """
        Select the vertices which are in front of the given plane.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        return self.select().vertices_in_front_of_plane(plane=plane).end()

    def keeping_vertices_on_or_behind_plane(self, plane):
        """
        Select the vertices which are either on or behind the given plane.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        return self.select().vertices_on_or_behind_plane(plane=plane).end()

    def keeping_vertices_behind_plane(self, plane):
        """
        Select the vertices which are behind the given plane.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        return self.select().vertices_behind_plane(plane=plane).end()

    def picking_vertices(self, indices_or_boolean_mask):
        """
        Select only the given vertices.

        Return a new mesh, without mutating the callee.

        Args:
            indices_or_boolean_mask (np.arraylike): Either a list of vertex
                indices, or a boolean mask the same length as the vertex array.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return (
            self.select()
            .pick_vertices(indices_or_boolean_mask=indices_or_boolean_mask)
            .end()
        )

    def picking_faces(self, indices_or_boolean_mask):
        """
        Select only the given faces.

        Return a new mesh, without mutating the callee.

        Args:
            indices_or_boolean_mask (np.arraylike): Either a list of vertex
                indices, or a boolean mask the same length as the vertex array.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return (
            self.select()
            .pick_faces(indices_or_boolean_mask=indices_or_boolean_mask)
            .end()
        )

    def picking_face_groups(self, *group_names):
        """
        Select faces which belong to the given face groups.

        Args:
            group_names (list): The face groups to keep.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().pick_face_groups(*group_names).end()

    def sliced_by_plane(self, *planes, only_for_selection=None):
        """
        Slice the triangles, keeping the portion in front of the given plane.

        - Faces partially in front of the plane are sliced.
        - Faces fully in front of the plane are kept as is.
        - Faces fully behind the plane are culled.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.
            only_for_selection (function): A function which receives a
                `lacecore.Selection` and should invoke selection methods on it.

        Returns:
            lacecore.Mesh: The sliced mesh.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        from polliwog import Plane
        from polliwog.plane import slice_triangles_by_plane
        from .._mesh import Mesh

        for plane in planes:
            assert isinstance(plane, Plane)

        working = self
        for plane in planes:
            # Since slicing renumbers the faces, recompute `faces_to_slice`
            # after each slice.
            if only_for_selection is None:
                faces_to_slice = None
            else:
                selection = working.select()
                only_for_selection(selection)
                faces_to_slice, _ = selection.generate_masks()

            vertices, faces, face_mapping = slice_triangles_by_plane(
                vertices=working.v,
                faces=working.f,
                plane_reference_point=plane.reference_point,
                plane_normal=plane.normal,
                faces_to_slice=faces_to_slice,
                ret_face_mapping=True,
            )
            face_groups = (
                None
                if working.face_groups is None
                else working.face_groups.reindexed(face_mapping)
            )
            working = Mesh(v=vertices, f=faces, face_groups=face_groups)
        return working



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lacecore',
 'lacecore._analysis',
 'lacecore._cli',
 'lacecore._common',
 'lacecore._obj',
 'lacecore._selection',
 'lacecore._transform',
 'lacecore.shapes']

package_data = \
{'': ['*']}

install_requires = \
['numpy', 'ounce>=1.1.0,<2.0', 'polliwog>=3.0.0a3', 'vg>=2.0.0']

extras_require = \
{'cli': ['click>=8.1.3,<9', 'tinymetabobjloader==2.0.0a0'],
 'obj': ['tinymetabobjloader==2.0.0a0']}

entry_points = \
{'console_scripts': ['convert_units = '
                     'lacecore_cli.convert_units:convert_units']}

setup_kwargs = {
    'name': 'lacecore',
    'version': '3.0.0a3',
    'description': 'Polygonal meshes optimized for cloud computation',
    'long_description': 'None',
    'author': 'Paul Melnikow',
    'author_email': 'github@paulmelnikow.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/lace/lacecore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
